#!/usr/bin/env python2
# * **************************************************************** **
# File: audio.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

""" Module providing conversions to, and analysis of, wav files """

from __future__ import print_function
from __future__ import absolute_import

import wave
import tempfile
import os.path
import sys

DOWNLOAD_BYTES_BETWEEN_PAUSES = 1536
DOWNLOAD_PAUSE_MSECS = 2000

WAVE_SAMPLE_RATE_HZ = 44100

# A ramping function between two different samples - the
# values are the percent of the change to apply in each sample.
# values supplied by Brenton (24/Jan/16)
RAMP = (1, 3, 7, 16, 50, 84, 93, 97, 99)

# A quanta is a 1/2 a microsecond. As the sample rate is in
# Hz, we have to divide it by 2000 to get samples per 0.5ms.
SAMPLES_PER_QUANTA = WAVE_SAMPLE_RATE_HZ / 2000

PULSE_AUDIO = True

# ############ main audio creator class ###############################################

# i2b is function for converting int to byte
if sys.version_info[0] == 2:
    i2b = chr
else:
    i2b = lambda x: bytes([x])

class Output(object):
    """Create a wav file"""

    def __init__(self, dir, nameOverride=None):
        """Create an audio file within a directory (typically creating a new name)"""
        self.directory = dir
        if nameOverride:
            self.filename = nameOverride
            self.fileHandle = open(self.filename, "wb")
        else:
            self.fileHandle = tempfile.NamedTemporaryFile(mode="wb",
                                                          prefix="tok", suffix=".wav",
                                                          dir=self.directory, delete=False)
            self.filename = self.fileHandle.name

        self.sampleRate = 44100
        self.samplesPerQuanta = self.sampleRate / 2000
        self.lastLeft = 128
        self.lastRight = 128
        self.downloadBytesBetweenPauses = 1536
        self.downloadPauseMsecs = 2000

        if (PULSE_AUDIO):
            self.audio_func = self.createAudioWithPulses
            self.silence_func = self.createSilenceWithPulses
        else:
            self.audio_func = self.createAudioRamping
            self.silence_func = self.createSilenceRamping


    def SetSampleRate(self, sampleRate):
        self.sampleRate = sampleRate
        self.samplesPerQuanta = self.sampleRate / 2000

    def GetWavPath(self):
        return os.path.join(self.directory, self.filename)

    def CreateDebugWav(self):
        waveWriter = wave.open(self.fileHandle)
        waveWriter.setnchannels(2)
        waveWriter.setsampwidth(1)
        waveWriter.setframerate(self.sampleRate)
        waveWriter.setcomptype("NONE", "")

        # now generate the test file
        data = chr(255) + chr(0) + \
            chr(128) + chr(128) + \
            chr(0) + chr(255) + \
            chr(128) + chr(128)
        count = 2000
        while count > 0:
            waveWriter.writeframes(data)
            count -= 1

        waveWriter.close()

    # def WriteProgramWav(self, binaryString):
    #     self.WriteWav(TOKEN_DOWNLOAD_STR + TOKEN_VERSION_STR + binaryString)

    # def WriteFirmwareWav(self, binaryString):
    #     self.WriteWav(FIRMWARE_DOWNLOAD_STR + FIRMWARE_VERSION_STR + binaryString)

    def WriteWav(self, binaryData):
        waveWriter = wave.open(self.fileHandle)
        waveWriter.setnchannels(2)
        waveWriter.setsampwidth(1)
        waveWriter.setframerate(self.sampleRate)
        waveWriter.setcomptype("NONE", "")

        self.lastLeft = 128
        self.lastRight = 128
        self.ConvertWithPause(binaryData, waveWriter)
        waveWriter.close()

    def ConvertWithPause(self, binString, waveWriter):
        index = 0
        preamble = 0
        pauseCount = 0

        # 500 milliseconds (1000 midQuantas) of silence at the beginning
        waveWriter.writeframes(self.silence_func(1000, self.sampleRate))

        preamble = 0
        while (preamble < self.samplesPerQuanta):
            waveWriter.writeframes(self.audio_func(0, self.sampleRate))
            preamble += 1

        while (index < len(binString)):
            if (pauseCount == self.downloadBytesBetweenPauses):
                preamble = 0
                while (preamble < self.downloadPauseMsecs):
                    waveWriter.writeframes(self.audio_func(0, self.sampleRate))
                    preamble += 1
                pauseCount = 0

            data = binString[index]

            # start
            waveWriter.writeframes(self.audio_func(6, self.sampleRate))

            # now the actual data -- big endian or little endian
            mask = 1
            ones = 0
            while (mask <= 0x80):
                if (data & mask):
                    waveWriter.writeframes(self.audio_func(2, self.sampleRate))
                    ones += 1
                else:
                    waveWriter.writeframes(self.audio_func(0, self.sampleRate))
                mask <<= 1

            # add stop - BBB Changed to 8 - differs from start
            waveWriter.writeframes(self.audio_func(8, self.sampleRate))

            index += 1
            pauseCount += 1

        # added to end as well - to ensure entire data is played. - ## BBB
        preamble = 0
        while (preamble < self.samplesPerQuanta):
            waveWriter.writeframes(self.audio_func(0, self.sampleRate))
            preamble += 1

        # 500 milliseconds (1000 midQuantas) of silence at the end
        waveWriter.writeframes(self.silence_func(1000, self.sampleRate))

    def createAudioRamping(self, midQuantas, sample_rate):
        data = b""
        samples_per_quanta = sample_rate / 2000

        # write fars
        data += self.ramp(255, 0, samples_per_quanta)

        # write nears
        data += self.ramp(0, 255, samples_per_quanta)

        if (midQuantas > 0):
            data += self.ramp(128, 128, midQuantas * samples_per_quanta)

        return data

    def createAudioWithPulses(self, midQuantas, sample_rate):
        data = b""
        samples_per_quanta = sample_rate / 2000
        total_samples = 2 * samples_per_quanta + (midQuantas * samples_per_quanta)

        # write far
        data += i2b(255) + i2b(0)
        # write near
        data += i2b(0) + i2b(255)

        count = 2
        while count < total_samples:
            data += i2b(128) + i2b(128)
            count += 1

        return data

    def createSilenceRamping(self, midQuantas, sample_rate):
        samples_per_quanta = sample_rate / 2000
        return self.ramp(128, 128, midQuantas * samples_per_quanta)

    def createSilenceWithPulses(self, midQuantas, sample_rate):
        data = b""
        samples_per_quanta = sample_rate / 2000
        total_samples = midQuantas * samples_per_quanta

        count = 0
        while count < total_samples:
            data += i2b(128) + i2b(128)
            count += 1

        return data

    def ramp(self, newLeft, newRight, samples):
        # print "ramp", samples
        data = b""

        if (samples < len(RAMP)):
            print("ERROR - audio transition is smaller then the ramp size")
            sys.exit(1)

        diffLeft = newLeft - self.lastLeft
        diffRight = newRight - self.lastRight
        count = 0

        while (count < len(RAMP)):
            left = int(self.lastLeft + (diffLeft * RAMP[count] / 100))
            right = int(self.lastRight + (diffRight * RAMP[count] / 100))
            # print "Ramp %d/%d" % (left, right)
            data += i2b(left) + i2b(right)

            count += 1

        while (count < samples):
            # print "Stable %d/%d" % (newLeft, newRight)
            data += i2b(newLeft) + i2b(newRight)
            count += 1

        self.lastLeft = newLeft
        self.lastRight = newRight
        return data



#!/usr/bin/env python2
# * **************************************************************** **
# File: util.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

""" Module providing utility functions. """

from __future__ import print_function
from __future__ import absolute_import

import sys
import datetime
import os
import os.path


class Enum(object):
    """ Provides a 'C'-like enumeration for python
        e.g. ERRORS = Enum("OVERFLOW", "DIV_BY_ZERO")
    """

    def __init__(self, *keys):
        self.__dict__.update(zip(keys, range(len(keys))))

    def len(self):
        return len(self.__dict__.keys())

    def isValid(self, value):
        return (value >= 0) and (value < self.len())


class Mask(object):
    """ Provides a 'C'-like enumeration of mask values for python
        (so values are powers of 2, so they can be orred together)
        e.g. DUMP = Mask("PARSER", "OPTIMISER")
    """

    def __init__(self, *keys):
        values = [2**i for i in range(len(keys))]
        self.__dict__.update(zip(keys, values))

    def len(self):
        return len(self.__dict__.keys())

    def isValid(self, value):
        """value can be an 'or' of possible keys"""
        return (value >= 0) and (value < (2 ** self.len()))


class SimpleLog(object):
    """ Provide a VERY SIMPLE log of execution. Just start, and end with timestamps. So
        if something crashes then should be able to see that it happened."""

    def __init__(self, use=True, fileName="EdPy.log", maxBytes=2000000):
        self.fh = None
        self.use = use
        self.start = datetime.datetime.now()
        self.fileName = fileName
        self.maxBytes = maxBytes

    def formatTimestamp(self, ts=None):
        if (ts is None):
            return datetime.datetime.now.isoformat(' ')
        else:
            return ts.isoformat(' ')

    def formatDelta(self, delta):
        return "+{:d}.{:06d}s".format(delta.seconds, delta.microseconds)

    def open(self):
        if (not self.use):
            return

        # try to rename if it's too large. But if there is an error then just
        # continue on.
        try:
            # if there is an existing file then see if it's too large
            if (os.path.exists(self.fileName)):
                bytes = os.path.getsize(self.fileName)
                if (bytes >= self.maxBytes):
                    os.rename(self.fileName, self.fileName + ".old")
        except:
            pass

        try:
            # Non-buffered appending log
            self.fh = open(self.fileName, "a", buffering=0)
        except:
            self.fh = None

    def log(self, line):
        if (not self.use):
            return

        if (self.fh is None):
            self.open()

        if (self.fh is not None):
            now = datetime.datetime.now()
            delta = now - self.start
            print("{:s} dur:{:s} pid:{} msg:{:s}".format(self.formatTimestamp(now),
                                                         self.formatDelta(delta),
                                                         os.getpid(), line), file=self.fh)
            self.fh.flush()

    def close(self):
        if (not self.use):
            return

        if (self.fh is not None):
            self.fh.close()
            self.fh = None


def LowerStr(inString):
    """Returns a lower case string from any string. Useful for argparse types"""
    return inString.lower()


def CheckPythonVersion():
    ver = sys.version_info

    # Using version 2.0 access to version (instead of assuming 2.7 here)
    if (ver < (2,7)):
        rawText = "Python version must be 2.7 or greater,"
        rawText += " this Python version is %d.%d." % (ver[0], ver[1])
        print("FATAL: " + rawText, file=sys.stderr)
        sys.exit(1)

# do this check on every import of util!
CheckPythonVersion()



#!/usr/bin/env python


# * **************************************************************** **
#
# File: hl_parser.py
# Desc: high-level language parsing for the token assembler
# Note:
#
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
#
# Copyright 2006, 2014, 2016, 2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

from __future__ import print_function
from __future__ import absolute_import

import string
import copy
from . import io


device_types = {  # 'not used':0,
    # 'digin': 1, 'digout': 2, 'analogin': 3, 'bumper': 0xb,
    'tracker': 4, 'irtx': 5, 'irrx': 6, 'beeper': 7,
    'motor-a': 8, 'motor-b': 9, 'led': 0xa}

# Amount of ram to store the registers and work space for the interpreter
device_storage = {
    # 'digin': 4 + 0, 'digout': 5 + 0, 'analogin': 2 + 0, 'bumper': 1 + 0,
    'tracker': 4 + 0, 'irtx': 2 + 0, 'irrx': 5 + 0, 'beeper': 7 + 0,
    'motor-a': 5 + 4, 'motor-b': 5 + 4, 'led': 4 + 0}

special_names = ['_index', '_devices', '_timers', '_cpu']
special_dtypes = ['index', 'devices', 'timers', 'cpu']


reset_devices = {'_index': ('index', 12), '_devices': ('devices', 13),
                 '_timers': ('timers', 14), '_cpu': ('cpu', 15)}

reset_locations = {'c': ('index', 12), 'd': ('devices', 13),
                   'e': ('timers', 14), 'f': ('cpu', 15)}

devices = {}

locations = {}

registers = {
    # not used in edison
    # 'digin': {'status': (0, 1), 'action': (1, 1), 'pulsetime': (2, 2)},
    # 'digout': {'status': (0, 1), 'action': (1, 1), 'output': (2, 1),
    #            'pulsetime': (3, 2)},
    # 'analogin': {'level': (0, 2)},
    # 'bumper': {'status': (0, 1)},

    'tracker': {'status': (0, 1), 'output': (1, 1), 'lightlevel': (2, 2)},

    'led': {'status': (0, 1), 'output': (1, 1), 'lightlevel': (2, 2)},

    'beeper': {'status': (0, 1), 'action': (1, 1), 'freq': (2, 2),
               'duration': (4, 2),
               'tunecode': (6, 1),   # this used to be 'tune'
               'tunestring': (7, 1),
               'tempo': (8, 2)},

    'irtx': {'action': (0, 1), 'char': (1, 1)},

    'irrx': {'status': (0, 1), 'action': (1, 1), 'check': (2, 1),
             'match': (3, 1), 'char': (4, 1)},

    'motor-a': {'status': (0, 1), 'control': (1, 1), 'distance': (2, 2)},
    'motor-b': {'status': (0, 1), 'control': (1, 1), 'distance': (2, 2)},


    'index': {'action': (0, 1), '8b1cursor': (1, 1), '8b1step': (2, 1),
              '8b1window': (3, 1), '8b2cursor': (4, 1), '8b2step': (5, 1),
              '16b1cursor': (6, 1), '16b1step': (7, 1),
              '16b1window': (8, 2), '16b2cursor': (10, 1), '16b2step': (11, 1),
              '16b1temp': (12, 2), '16b2temp': (14, 2),

              # alternate meanings for EdPy.py unused registers
              '8b1local': (2, 1), '16b1local': (4, 2),
              '8b2local': (7, 1), '16b2local': (10, 2), },


    # Some not used in edison
    # 'devices': {'status': (0, 1), 'action': (1, 1), 'lcdaction': (2, 1),
    #             'serset1': (3, 1), 'serset2': (4, 1), 'serrx': (5, 1),
    #             'sertx': (6, 1), 'lcdbyte': (7, 1), 'lcdrp': (8, 1),
    #             'lcdcp': (9, 1), 'lcdword': (10, 2), 'random': (12, 1),
    #             'button': (13, 1)},

    'devices': {'status': (0, 1), 'random': (1, 1), },

    'timers': {'status': (0, 1), 'action': (1, 1), 'pause': (2, 2),
               'oneshot': (4, 2), 'system': (6, 4)},

    'cpu': {'acc': (0, 2), 'flags': (2, 1), 'counter': (3, 1),
            'pc': (4, 2), 'sp': (6, 2)},
}


def reset_devices_and_locations():
    global devices
    global locations
    devices.clear()
    locations.clear()
    devices = copy.deepcopy(reset_devices)
    locations = copy.deepcopy(reset_locations)


def AsmError_NO_RET(number, internalError=None, line=0):
    if (internalError):
        io.Out.ErrorRaw(internalError)

    io.Out.Error(io.TS.ASM_INTERNAL_ERROR, "file::: Assembler internal error {0}", number)
    raise program.AssemblerError


def dump_reg_help():
    """Dump locations, type names and register names as a help to programmers"""
    print("\nLocations:")
    print("  0-b : connectable module (use hex digit or name from DEVICE statement)")
    for i in 'cdef':
        name = locations[i][0]
        print("  %s   : special module %s (use hex digit or '_%s')" % (i, name, name))

    print("\nDevice types (prefixed with device code):")
    olist = device_types.items()
    olist.sort(key=lambda p: p[1])
    count = 0
    print("  ", end='')
    for name, num in olist:
        print("%d:%-10s " % (num, name), end='')
        count += 1
        if (count >= 5):
            print("\n  ", end='')
            count = 0
    if (count != 0):
        print()

    print("\nRegister names (offset/bytes):")
    rorder = [name for (name, num) in olist]
    rorder.extend(special_dtypes)
    # print(rorder)

    for r in rorder:
        reg_dict = registers[r]
        rlist = reg_dict.items()
        rlist.sort(key=lambda p: p[1][0])
        # print(rlist)
        if (r in special_dtypes):
            print("  _%s" % (r), end='')
        else:
            print("  %s" % (r), end='')
        count = 10
        for (name, (offset, size)) in rlist:
            count += 1
            if count >= 4:
                print("\n    ", end='')
                count = 0
            desc = "%s:(%s/%d)" % (name, string.hexdigits[offset], size),
            print("%-16s" % (desc), end='')

        print("\n")


def add_device(loc, dtype, name=None):
    global devices
    global locations

    if (dtype not in device_types):
        AsmError_NO_RET(200, "Unknown device type: " + dtype)
        return False
    if (loc < 0 or loc > 11):
        AsmError_NO_RET(201, "Location (%d) outside of 0 to 11" % (loc))
        return False

    # location already used?
    loc_hex = string.hexdigits[loc]
    if (loc_hex in locations):
        AsmError_NO_RET(202, "Location (%d) is already used with a unit of type: " % (loc) +
                        locations[loc_hex][0])
        return False

    if (name):
        name = name.lower()

    if (name in special_names):
        AsmError_NO_RET(203, "DEVICE name can not be one of the built-in names " + str(special_names))
        return False
    elif (name in devices):
        AsmError_NO_RET(204, "DEVICE name already used: " + name)
        return False

    # check the special ones - tracker and motors
    if (dtype == 'tracker' and loc != 0):
        AsmError_NO_RET(205, "A tracker device can only be located at 0, not location %d" % (loc))
        return False

    # motors need two spots
    if (dtype == 'motor-a' or dtype == 'motor-b'):
        if (loc == 11):
            motor_second_loc = 0
        else:
            motor_second_loc = loc + 1
        motor_second_hex = string.hexdigits[motor_second_loc]
        if (motor_second_hex in locations):
            AsmError_NO_RET(206, "Motors require two locations yet the second location is already used with a " +
                            locations[motor_second_hex][0])
            return False
        locations[motor_second_hex] = ('motor_second', motor_second_loc)

    locations[loc_hex] = (dtype, loc)

    # does it have a name?
    if (name):
        devices[name] = (dtype, loc)

    return True


def get_location_type_and_size(location):
    loc_hex = string.hexdigits[location]
    if (loc_hex in locations and locations[loc_hex][0] != 'motor_second'):
        dtype = locations[loc_hex][0]
        return (device_types[dtype], device_storage[dtype])
    else:
        return (0, 0)                   # Not used


def dump_devices():
    print("\nDevice mappings:")
    for i in range(12):
        loc_hex = string.hexdigits[i]
        if (loc_hex in locations):
            print("  %s - type:%s" % (loc_hex, locations[loc_hex][0]), end='')
            for d in devices:
                if (devices[d][1] == i):
                    print(", name:%s" % (d), end='')
                    break
            print()


class word(object):
    def __init__(self, type, value):
        self.type_ = type
        self.value_ = value

    def type(self):
        return self.type_

    def val(self):
        return self.value_

    def num(self):
        return parse_bases(self.value_)

    def anum(self):
        if (self.type_ not in ["arg", "const"]):
            AsmError_NO_RET(207, "Expected an argument or constant!")
            return 0

        if (self.type_ == "const"):
            return self.value_
        else:
            return parse_bases(self.value_)

    def astr(self):
        if (self.type_ not in ["arg", "string"]):
            AsmError_NO_RET(208, "Expected an argument or string!")
            return ""
        elif (len(self.value_) > 0):
            if (self.value_[0] not in string.ascii_letters + '_'):
                AsmError_NO_RET(209, "Word should have been a valid name!")
                return ""
            else:
                return self.value_

    def amodreg(self):
        if (self.type_ not in ["arg", "modreg"]):
            AsmError_NO_RET(210, "Expected an argument or string!")
            return ""

        elif (self.type_ == "modreg"):
            return self.value_
        else:
            return parse_mod_reg(self.value_)

    def __str__(self):
        if (isinstance(self.value_, int)):
            return "{Word: %s,0x%x}" % (self.type_, self.value_)
        else:
            return "{Word: %s,%s}" % (self.type_, self.value_)


def format_word_list(wlist):
    output = "["
    for w in wlist:
        output += str(w) + ", "
    if (wlist):
        output = output[:-2]
    output += "]"
    return output


# local utility functions

def parse_bases(snum, string=False):
    """detect the base and return a decimal equivalent"""
    if (snum.startswith("'") and snum.endswith("'")):
        # must be one ascii character
        num = ord(snum[1:-1])
    elif (string and snum.startswith('"') and snum.endswith('"')):
        AsmError_NO_RET(211, "Error - use single quotes for an ascii constant")
        return
    elif ((snum.endswith("/2"))):
        num = int(snum[:-2], 2)
    elif ((snum.endswith("/10"))):
        num = int(snum[:-3], 10)
    elif ((snum.endswith("/16"))):
        num = int(snum[:-3], 16)
    else:
        num = int(snum, 0)

    return num


def parse_mod_reg(smodreg):
    # print("parse_mod_reg():", smodreg)
    # dump_devices()
    if ((len(smodreg) == 2) and
        (smodreg[0] in string.hexdigits) and
        (smodreg[1] in string.hexdigits)):
        num = int(smodreg, 16)
    else:
        if (':' in smodreg):
            part = smodreg.lower().split(':', 2)
            if (len(part) != 2):
                AsmError_NO_RET(212, "Invalid mod/reg syntax: " + smodreg)
                return 0
            if (part[0] in devices):
                dtype = devices[part[0]][0]
                loc = devices[part[0]][1]
            elif (part[0] in locations):
                dtype = locations[part[0]][0]
                loc = locations[part[0]][1]
            else:
                AsmError_NO_RET(213, "Didn't understand the module name/number: " + part[0])
                return 0

            # now the register name/number
            regs = registers[dtype]
            if (part[1] in regs):
                num = regs[part[1]][0] + (loc << 4)
            elif (part[1] in '0123456789abcdef'):
                num = int(part[1], 16) + (loc << 4)
            else:
                AsmError_NO_RET(214, "Didn't understand the register name/number: " + part[1])
                return 0

        else:
            AsmError_NO_RET(215, "Error - can't parse %s as a modreg" % (smodreg))
            return

    # print("modreg: %s -> %02x" % (smodreg, num))
    return num


def prechop_line(line):
    """Handle strings and comments"""
    in_string = ''
    in_escape = False
    components = []
    comp = ""

    for c in line:
        if (c == '\n'):
            break

        if (in_escape):
            comp += c
            in_escape = False

        elif (in_string and c == '\\'):
            in_escape = True

        elif (c == '#' and not in_string):
            # comment from here to end of line - bail out now
            break

        elif (not in_string and c in ['"']):
            in_string = c
            comp += c

        elif (in_string and c == in_string):
            in_string = ''
            comp += c

        elif (not in_string and c in [' ', '\t', ',']):
            if (comp):
                components.append(comp)
            comp = ''

        else:
            comp += c

    if (comp):
        components.append(comp)

    return components


def chop_line(line):
    """Filter out comments then create words from the line"""

#    segs_spaces = line.split()
#    segs = []
#    for ss in segs_spaces:
#        segs.extend(ss.split(","))

    segs = prechop_line(line)
    # print(segs)

    words = []
    for s in segs:
        # skip empty segs
        if not s:
            continue

        # terminate at a comment
        if (s.startswith('#')):
            break

        # lines can start with a label or operator
        if (s.startswith(":")):
            # a jump label
            words.append(word("label", s[1:]))

        elif (not words):
            # first one so must be the operator
            words.append(word("op", s))

        elif ((s.startswith('"') and s.endswith('"')) or
              (s.startswith("'") and s.endswith("'"))):
            if (len(s) <= 2):
                continue
            else:
                words.append(word("string", s[1:-1]))

        elif (s.startswith("$")):
            # should be a constant
            if (len(s) == 4 and s[1] == "'" and s[3] == "'"):
                num = ord(s[2])
            else:
                num = parse_bases(s[1:])
            words.append(word("const", num))

        elif (s.startswith("%")):
            # a module/register
            modreg = parse_mod_reg(s[1:])
            words.append(word("modreg", modreg))

        elif (s.startswith(":")):
            # a jump label
            words.append(word("label", s[1:]))

        elif (s.startswith("@")):
            # a variable name - add another element to the word
            words.append(word("var", s[1:]))

        else:
            words.append(word("arg", s))

    # print(words)
    return words



#!/usr/bin/env python2
# * **************************************************************** **
# File: token_bits.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

"""Contains definitions for bits in the token modules. Typically
   STATUS bits are masks, and CONTROL bits are bit numbers (to be
   used with setbit)
"""
STATUS_LINE_OVER_LINE = 1
STATUS_LINE_CHANGED = 2
CONTROL_LINE_POWER = 1

CONTROL_LED_POWER = 1

STATUS_MOTOR_STRAIN = 1
STATUS_MOTOR_DISTANCE = 2
CONTROL_MOTOR_SPEED_MASK = 0x0f
CONTROL_MOTOR_REMAIN_GOING = 4
CONTROL_MOTOR_CMD_MASK = 0xe0

STATUS_IRX_BILL_RECEIVED = 0x01
STATUS_IRX_MATCHED =       0x02
STATUX_IRX_CHECK_VALID =   0x04
STATUS_IRX_OBS_RIGHT =     0x08
STATUS_IRX_OBS_CENTRE =    0x10
STATUS_IRX_OBS_LEFT =      0x20
STATUS_IRX_OBS_DETECTED =  0x40
CONTROL_IRX_DO_CHECK = 0

STATUS_BEEP_TUNE_DONE =    0x01
STATUS_BEEP_TONE_DONE =    0x02
STATUS_BEEP_CLAP_DETECTED = 0x04
STATUS_BEEP_TS_ERROR =     0x08
CONTROL_BEEP_PLAY_CODED_TUNE =   0
CONTROL_BEEP_PLAY_TONE =         1
CONTROL_BEEP_PLAY_BEEP =         2
CONTROL_BEEP_PLAY_STRING_TUNE =  3

CONTROL_ITX_TRANSMIT_CHAR =      0
CONTROL_ITX_DO_OBSTACLE_DETECT = 1

CONTROL_INDEX_WRITE_16BIT = 1
CONTROL_INDEX_READ_16BIT =  2
CONTROL_INDEX_WRITE_8BIT =  5
CONTROL_INDEX_READ_8BIT =   6

STATUS_DEVICE_BUTTON_1 = 0x08
STATUS_DEVICE_BUTTON_2 = 0x04
STATUS_DEVICE_BUTTON_3 = 0x02
STATUS_DEVICE_BUTTON_4 = 0x01

STATUS_TIMER_ONE_SHOT_EXPIRED = 0x01
STATUS_TIMER_ONE_SHOT_RUNNING = 0x02

CONTROL_TIMER_TRIGGER_ONE_SHOT = 0
CONTROL_TIMER_TRIGGER_PAUSE =    1
CONTROL_TIMER_ENABLE_SLEEP =     2



import os.path
from thonny import get_workbench, get_runner
from thonny.ui_utils import select_sequence


def program_edison_enabled():
    code = _get_current_code()
    return code is not None and code.strip() != ""

def _get_current_code():
    editor = get_workbench().get_editor_notebook().get_current_editor()
    if editor is None:
        return None
    return editor.get_code_view().text.get("1.0", "end")

                    
def program_edison():
    get_runner().execute_current("ProgramEdison")

def load_plugin():
    image_path = os.path.join(os.path.dirname(__file__), "res", "tools.program_edison.png")
    get_workbench().add_command("program_edison", "tools", "Send current script to Edison",
                                program_edison,
                                program_edison_enabled,
                                default_sequence=select_sequence("<Control-e>", "<Command-e>"),
                                group=120,
                                image=image_path,
                                caption="Program Edison",
                                include_in_toolbar=True)



#!/usr/bin/env python2
# * **************************************************************** **
# File: io.py
# Requires: Python 2.7+ (but not Python 3.0+)
# Note: For history, changes and dates for this file, consult git.
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
# Copyright 2015-2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

""" Module providing input/output support including an interface to translated strings. """

from __future__ import print_function
from __future__ import absolute_import

import json
import re
import sys
import traceback
import types

from . import util

# Translation string enumeration - returns a number
TS = util.Enum("ELPY_SPECIAL_FAIL",
               "TOP_PREFIX", "ERROR_PREFIX", "WARN_PREFIX", "INFO_PREFIX", "VERBOSE_PREFIX", "DEBUG_PREFIX",
               "FILE_OPEN_ERROR", "BAD_INPUT_CHARS",

               "PARSE_START", "PARSE_ERROR", "PARSE_NAME_REUSED",
               "PARSE_INVALID_STATEMENT", "PARSE_GLOBAL_ORDER",
               "PARSE_IMPORT_NOT_ED", "PARSE_IMPORT_ORDER",
               "PARSE_SYNTAX_ERROR", "PARSE_NOT_SUPPORTED", "PARSE_TOO_COMPLEX",
               "PARSE_NOT_IN_LOOP", "PARSE_CLASS_ARG0_NOT_SELF",
               "PARSE_MIXED_RETURNS", "PARSE_CLASS_ALL_STATEMENTS_IN_FUNCTIONS",
               "PARSE_CLASS_NO_BASES_ALLOWED", "PARSE_CONST_NOT_INT",
               "OPT_START", "OPT_RESERVED_NAME",
               "OPT_INCORRECT_ARG_USE", "OPT_INCORRECT_ARG_DEFINE",
               "OPT_VAR_NOT_BOUND", "OPT_VAR_NOT_INT", "OPT_VAR_TYPE_CHANGED",
               "OPT_STRING_NOT_ALLOWED", "OPT_LIST_NOT_ALLOWED",
               "OPT_CLASS_INIT_ERROR", "OPT_CLASS_DATA_ERROR",
               "OPT_NOT_CLASS_REF", "OPT_NOT_ASSIGNABLE",
               "OPT_SLICE_NOT_ALLOWED", "OPT_LCL_HIDES_GLB",
               "OPT_NOT_A_GLOBAL_VAR", "OPT_WRITE_TO_ED_PY_CONSTANT",
               "OPT_FUNCTION_NOT_DEFINED", "OPT_NOT_SUPPORTED",
               "OPT_VAR_MUST_BE_TS_OR_LIST", "OPT_ONLY_AT_TOP_LEVEL",
               "OPT_SELF_NOT_IN_METHOD",
               "OPT_UNKNOWN_FUNCTION", "OPT_UNKNOWN_ED_FUNCTION",
               "OPT_MISSING_ED_IMPORT",
               "OPT_CONSTANT_TOO_NEGATIVE", "OPT_CONSTANT_TOO_POSITIVE",
               "OPT_ED_ASSIGN_NOT_CONSTANT", "OPT_ED_ASSIGN_AGAIN",
               "OPT_ED_ASSIGN_BAD_VALUE", "OPT_ED_ASSIGN_NOT_SET",
               "OPT_ED_ASSIGN_IN_FUNCTION", "OPT_ED_FUNCTION_NOT_AVAILABLE",
               "OPT_ED_FUNCTION_NOT_USEFUL", "OPT_ED_WARN_TUNESTRING_END",
               "OPT_ED_LIST_TOO_LONG", "OPT_BAD_EVENT_NUMBER",

               "CMP_START", "CMP_INTERNAL_ERROR", "CMP_VAR_NOT_BOUND",
               "ASM_START", "ASM_MEM_OVERFLOW", "ASM_INTERNAL_ERROR",
)

# Output level
LEVEL = util.Enum("ERROR", "WARN", "TOP", "INFO", "VERBOSE", "DEBUG")

# SINK.CONSOLE is to stdout/stderr, SINK.JSON is to the JSON object,
# SINK.TEST outputs the number instead of the text to make comparisons easier
SINK = util.Enum("CONSOLE", "JSON", "BOTH", "TEST")

# DUMP mask to signal that different datastructures should be dumped
DUMP = util.Mask("PARSER", "OPTIMISER", "COMPILER", "ASSEMBLY", "BINARY")

# ############ Json output class ###############################################


class JsonOutput(object):
    """Accumulate output for the JSON output sink."""

    def __init__(self):
        self.error = False
        self.messages = []
        self.wavFilename = None

    def Out(self, level, message):
        if (level == LEVEL.ERROR):
            self.error = True
        self.messages.append(message)  # message already has the level encoded in it

    def ForceError(self, error):
        self.error = error

    def SetWavFilename(self, wavFilename):
        self.wavFilename = wavFilename

    def Convert(self):
        structure = {
            "error": self.error,
            "messages": self.messages,
            "wavFilename": self.wavFilename
        }
        return json.JSONEncoder().encode(structure)


# ############ Output class ###############################################


class OutClass(object):
    def __init__(self):
        self.outputSink = SINK.CONSOLE
        self.maxOutputLevel = LEVEL.VERBOSE
        self.reRaise = False
        self.dumpMask = 0
        self.langFileHandle = None
        self.jsonOutput = JsonOutput()
        self.errorRaised = False
        self.outputString = ""

        self.errorRawContextLevel = 0
        self.errorRawContext = ["", "", "", "", "", "", "", "", "", ""]

        self.rawPrefix = [
            (TS.ERROR_PREFIX, "ERR"),
            (TS.WARN_PREFIX, "WARN"),
            (TS.TOP_PREFIX, "TOP"),
            (TS.INFO_PREFIX, "INFO"),
            (TS.VERBOSE_PREFIX, "VERB"),
            (TS.DEBUG_PREFIX, "DBG")
        ]

        self.transPrefix = []
        for num, msg in self.rawPrefix:
            # TODO: Do the actual translation
            self.transPrefix.append(msg)

    def SetSink(self, sink):
        if (SINK.isValid(sink)):
            self.outputSink = sink
        else:
            self.FatalRaw("Invalid util.Enum SINK constant")

    def SetMaxLevel(self, level):
        if (LEVEL.isValid(level)):
            self.maxOutputLevel = level
        else:
            self.FatalRaw("Invalid util.Enum LEVEL constant")

    def SetLangFileHandle(self, langFileHandle):
        self.langFileHandle = langFileHandle

    def SetReRaise(self, reRaise):
        self.reRaise = reRaise

    def IsReRaiseSet(self):
        return self.reRaise

    def WasErrorRaised(self):
        return self.errorRaised

    def GetOutputAsString(self):
        return self.outputString

    def SetInfoDumpMask(self, dumpMask):
        if (DUMP.isValid(dumpMask)):
            self.dumpMask = dumpMask
        else:
            self.FatalRaw("Invalid util.Mask DUMP mask")

    def GetInfoDumpMask(self):
        return self.dumpMask

    def ForceJsonError(self, error):
        self.jsonOutput.ForceError(error)

    def SetWavFilename(self, wavFilename):
        self.jsonOutput.SetWavFilename(wavFilename)

    def Flush(self):
        if (self.outputSink == SINK.BOTH or self.outputSink == SINK.JSON):
            print(self.jsonOutput.Convert())

    def FatalRaw(self, rawText):
        """Output a fatal error without translation (so it can
           be used on translation errors!) to the stderr AND configured sync.
        """
        print("FATAL: " + rawText, file=sys.stderr)
        print("----------------\nStack Trace:")
        stack = traceback.format_stack()
        for l in stack[:-1]:
            print(repr(l))
        sys.exit(1)

    def DebugRaw(self, rawText, *args):
        """Output debug information without translation if the sink is CONSOLE.
        """
        if ((self.maxOutputLevel >= LEVEL.DEBUG) and
            (self.outputSink == SINK.CONSOLE or self.outputSink == SINK.BOTH)):
            if (not args):
                print("**DebugRaw**:", rawText)
            else:
                print("**DebugRaw**:", rawText, *args)

    def SetErrorRawContext(self, level, rawText):
        if (level > 0 and level < 10):
            self.errorRawContextLevel = level
            self.errorRawContext[level - 1] = rawText

    def ErrorRaw(self, rawText, *args):
        """Output error information without translation if the sink is CONSOLE.
        """
        self.errorRaised = True
        if ((self.maxOutputLevel >= LEVEL.ERROR) and
            (self.outputSink == SINK.CONSOLE or self.outputSink == SINK.BOTH)):

            for level in range(self.errorRawContextLevel):
                print("**ErrorRaw** - Context:", level, self.errorRawContext[level])

            if (not args):
                print("**ErrorRaw**:", rawText)
            else:
                print("**ErrorRaw**:", rawText, *args)

    def Top(self, stringNumber, rawText, *args):
        self.__Out(LEVEL.TOP, stringNumber, rawText, args)

    def Error(self, stringNumber, rawText, *args):
        self.errorRaised = True
        self.__Out(LEVEL.ERROR, stringNumber, rawText, args)

    def Warning(self, stringNumber, rawText, *args):
        self.__Out(LEVEL.WARN, stringNumber, rawText, args)

    def Info(self, stringNumber, rawText, *args):
        self.__Out(LEVEL.INFO, stringNumber, rawText, args)

    def Verbose(self, stringNumber, rawText, *args):
        self.__Out(LEVEL.VERBOSE, stringNumber, rawText, args)

    def __Out(self, outputLevel, stringNumber, rawText, args):
        """Output to sink, after translation, of text - replacing
           args (w/o translation) if they exist.
           Args in the text are denoted by {number[!.*][:.*]} as per the python
           Formatter syntax.
        """

        # Is this level desired to be output?
        if (outputLevel > self.maxOutputLevel):
            return

        # self.DebugRaw(outputLevel, stringNumber, rawText, len(args), args)

        # Count the number of pos arguments in the rawText. This must
        # equal the args (and the pos arguments in the translated text).
        potentialsRe = re.compile("\{(?!\{).*?\}")  # potential format markers
        potentials = re.findall(potentialsRe, rawText)
        # self.DebugRaw("potentials = {}".format(potentials))
        posCount = 0
        if (len(potentials) > 0):
            found = []
            pArgsRe = re.compile("\{(\d+)(?:[:!].*)?\}")

            for pot in potentials:
                match = re.match(pArgsRe, pot)
                # if match is None:
                #    self.FatalRaw("Bad syntax for translation string positional arguments.")

                # group 1 is the decimal pos number
                pos = int(match.group(1))
                # self.DebugRaw("match = {}".format(match.group(0)))
                if (pos not in found):
                    found.append(pos)

            # Ensure that there are no gaps in the positional arguments
            while len(found):
                if (posCount not in found):
                    self.FatalRaw("TransString positional arguments are not contiguous.")
                found.remove(posCount)
                posCount += 1

        # posCount is the number of positional arguments, so must match the length of args
        # self.DebugRaw("posCount = {}".format(posCount))
        if (posCount != len(args)):
            self.FatalRaw("TransString pos count {0} doesn't match the number of args {1}".format(posCount, len(args)))

        # TODO: Do the actual translation. For now just skipping that bit
        transText = self.transPrefix[outputLevel] + ": " + rawText
        # self.DebugRaw(transText + str(args))
        if (posCount):
            outText = transText.format(*args)
        else:
            outText = transText

        if (len(self.outputString) > 0):
            self.outputString += "{:s}|".format(outText)
        else:
            self.outputString = "|{:s}|".format(outText)

        if (self.outputSink == SINK.CONSOLE or self.outputSink == SINK.BOTH):
            print(outText)

        if (self.outputSink == SINK.JSON or self.outputSink == SINK.BOTH):
            self.jsonOutput.Out(outputLevel, outText)

        if (self.outputSink == SINK.TEST):
            print(outputLevel, ',', stringNumber, sep='', end='')
            if (posCount):
                for a in args:
                    print(',', a, sep='', end='')
                print()

    def DebugDumpObjectRaw(self, obj, name=''):

        if ((self.maxOutputLevel >= LEVEL.DEBUG) and
            (self.outputSink == SINK.CONSOLE or self.outputSink == SINK.BOTH)):

            filt = [types.MethodType]
            print("**DebugDump**: object %s (%r)" % (name, obj))
            dir_data = dir(obj)
            data = {}

            for d in dir_data:
                if ((d in ("__module__", "__doc__")) or
                    (not d.startswith("__"))):

                    temp = getattr(obj, d, None)
                    filter_out = False
                    if (temp is not None):
                        filter_out = False
                        for f in filt:
                            if (type(temp) is f):
                                filter_out = True
                                break

                        if (not filter_out):
                            data[d] = temp

            print("**DebugDump** -- %s" % (data))


# the singleton which everyone will use
Out = OutClass()



#!/usr/bin/env python

# * **************************************************************** **
#
# File: tokens.py
# Desc: Operation on token data structures
# Note:
#
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
#
# Copyright 2006, 2014, 2016, 2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

#
# Modified for the EdPy.py versions - starting at 0x60
#
# Token Version handled - (Major/Minor) = 0x6/0x0
# Firmware Version handled - (Major/Minor) = 0xe/0x0, 0x20/0x0
#

from __future__ import print_function
from __future__ import absolute_import

from . import io
from . import program

MIN_BYTE = 0
MAX_BYTE = 0xff
MIN_WORD = -0x7fff
MAX_WORD = 0x7fff
MAX_UWORD = 0xffff

MIN_SBYTE = -0x7f
MAX_SBYTE = 0x7f

# The firmware has a problem correctly computing the CRC when the size is a
# multiple of 256. This function finds all of the 'BAD' lengths so that
# the code can pad up by a byte and get around the firmware problem
def is_bad_length_for_crc(data_len):
    return (data_len >= 254) and ((data_len - 254) % 256 == 0)

# print("Test 254:", is_bad_length_for_crc(254))
# print("Test 255:", is_bad_length_for_crc(255))
# print("Test 509:", is_bad_length_for_crc(509))
# print("Test 510:", is_bad_length_for_crc(510))
# print("Test 511:", is_bad_length_for_crc(511))
# print("Test 765:", is_bad_length_for_crc(765))
# print("Test 766:", is_bad_length_for_crc(766))
# print("Test 767:", is_bad_length_for_crc(767))

# No LCD,  so lcd limit == 0
LIMIT_NAMES = ("Bytes", "Words", "LCD chars", "Event handlers", "Token bytes")
MAX_LIMITS = (256, 256, 0, 16, 4096)


space_types = {'b': 0, 'B': 0, 'w': 1, 'W': 1, 'a': 2, 'A': 2}
space_names = ("Byte", "Word", "LCD")


def AsmError_NO_RET(number, internalError=None, line=0):
    if (internalError):
        io.Out.ErrorRaw(internalError)

    io.Out.Error(io.TS.ASM_INTERNAL_ERROR, "file::: Assembler internal error {0}", number)
    raise program.AssemblerError


def calculate_crc(bytes):
    crc = 0xffff
    for i in range(len(bytes)):
        crc = crc ^ (bytes[i] << 8)
        for j in range(8):
            if (crc & 0x8000 != 0):
                crc = ((crc << 1) ^ 0x1021)
            else:
                crc = crc << 1

    return crc & 0xffff


def word_to_bytes(word):
    word = word & 0xffff
    return (((word >> 8) & 0xff), (word & 0xff))


class TokenStream(object):
    def __init__(self):
        self.clear()

    def clear(self):
        self.token_stream = []          # list of tokens in order
        self.sector_size = -1           # Whether to ack/nak and when
        self.limits = list(MAX_LIMITS)  # limits for the namespaces and tokens
        self.name_space = [[], [], []]  # Variables requested in namespaces
        self.devices = {}               # Devices indexed on location
        self.labels = {}                # Labels that we can jump to
        self.current_sections = []      # The section(s) that we are currently in
        self.section_breaks = []        # Recording where sections start and stop
        self.section_args = []          # ... and the args that were passed with it
        self.section_count = 0          # The number of the current section
        self.version = None             # The version number as (major, minor)
        self.download_type = None       # Can only have one type per download

    def add_token(self, token, explicit_placement=-1):
        spec_type = token.get_type()
        if (spec_type == "finish"):
            if (len(self.current_sections) > 0):
                AsmError_NO_RET(100, "FINISH must be after all sections.")
        elif (spec_type == "data"):
            if ((len(self.current_sections) == 0) or
                (self.current_sections[-1] not in ["main", "event"])):
                AsmError_NO_RET(101, "DATA must be in a 'main' or 'event' section.")

        elif (spec_type == "binary"):
            if ((len(self.current_sections) == 0) or
                (self.current_sections[-1] not in ["firmware"])):
                AsmError_NO_RET(102, "INSERT BINARY must be in a 'firmware' section.")
        else:
            if ((len(self.current_sections) == 0) or
                (self.current_sections[-1] not in ["main", "event"])):
                AsmError_NO_RET(103, "Tokens must be in a 'main' or 'event' section.")

        # Make jump labels local to the section
        jump = token.get_jump_label()
        if (jump):
            if (not jump[1].startswith(':')):
                # a local jump
                token.set_jump_label(jump[0], jump[1] + "_%d" % (self.section_count), jump[2])

        if (explicit_placement < 0):
            self.token_stream.append(token)
        else:
            if (explicit_placement >= len(self.token_stream)):
                AsmError_NO_RET(104, "Explicitly placing a token after the end of the stream.")
            else:
                self.token_stream.insert(explicit_placement, token)

    def add_token_in_fixups(self, token, explicit_placement):
        if (explicit_placement >= len(self.token_stream)):
            AsmError_NO_RET(105, "Explicitly placing a token after the end of the stream.")
        else:
            self.token_stream.insert(explicit_placement, token)

    def add_label(self, name):
        if ((len(self.current_sections) == 0) or
            (self.current_sections[-1] not in ["main", "event"])):
            AsmError_NO_RET(106, "Labels must be in a 'main' or 'event' section.")

        # Make the labels local to the section unless they start with '::'
        # (one ':' has been stripped off)
        if (name.startswith(':')):
            new_name = name
        else:
            new_name = "%s_%d" % (name, self.section_count)

        if (new_name in self.labels):
            AsmError_NO_RET(107, "Label %s defined twice in the same section." % (name))

        self.labels[new_name] = self.stream_marker()

    def stream_marker(self):
        return len(self.token_stream)

    # Define tighter limits on the name spaces
    # ARGS all numbers, ranges NOT checked
    def set_limits(self, b_lim, w_lim, l_lim, e_handlers, t_lim):
        if (self.section_count > 0):
            AsmError_NO_RET(108, "LIMITS must be before all sections.")

        new_limits = [b_lim, w_lim, l_lim, e_handlers, t_lim]
        for i in range(5):
            if (new_limits[i] < 0):
                AsmError_NO_RET(109, "Limit %s can't be less then 0." % (LIMIT_NAMES[i]))
            elif (new_limits[i] > MAX_LIMITS[i]):
                AsmError_NO_RET(110, "Limit %s too large." % (LIMIT_NAMES[i], MAX_LIMITS[i]))

        self.limits = new_limits

    def set_comms(self, sector_size):
        if (sector_size < 0):
            AsmError_NO_RET(111, "Comms sector_size can't be less then 0.")

        self.sector_size = sector_size

    def finish_tokens(self):
        """Absolute last token"""
        pass
        # io.Out.ErrorRaw("IMPLEMENT THIS FUNCTION.")

    # reserve space in the variable name spaces
    # ARGS: space 0-2, start & length are numbers
    def reserve_name_space(self, space, start, length):
        if (self.section_count > 0):
            AsmError_NO_RET(112, "RESERV must be before all sections.")

        if (start < 0 or length <= 0):
            AsmError_NO_RET(113, "Negative start or length is not allowed")

        if ((start + length) > self.limits[space]):
            io.Out.ErrorRaw("Reserve space extends beyond the max for %s space." % (space_names[space]))
            io.Out.Error(io.TS.ASM_MEM_OVERFLOW,
                         "file::: Overflowed {0} memory", space_names[space])
            raise program.AssemblerError


        self.name_space[space].append(("rsvd", "", start, length))

    def add_variable(self, space, name, start, length=0):
        io.Out.DebugRaw("Add_variable space:%s, name:%s, start:%d, length:%d" %
                        (space_names[space], name, start, length))

        if (name == '*'):
            name = ""

        if (length < 0):
            AsmError_NO_RET(114, "Negative length of data is not allowed")

        self.name_space[space].append(("data", name, start, length))

    # Add a device, location and optional name
    # ARGS_CHECKED: type is 0 to 15, location is 0 to 11, size is a byte
    def add_device(self, type, location, size):
        io.Out.DebugRaw("Add_device type:%d, loc:%d, size:%d" % (type, location, size))

        if (self.section_count > 0):
            AsmError_NO_RET(115, "DEVICE must be before all sections.")

        self.devices[location] = (type, size)

    def add_version(self, major, minor):
        io.Out.DebugRaw("Add_version major:%d, minor:%d" % (major, minor))

        if (self.section_count > 0):
            AsmError_NO_RET(116, "VERSION must be before all sections.")

        self.version = (major, minor)

    def add_begin(self, type, arg1=-1, arg2=-1, arg3=-1):
        """Start of a new section"""
        io.Out.DebugRaw("Add_begin type:%s, arg1:%d, arg2:%d, arg3:%d" % (type, arg1, arg2, arg3))

        # do checks to see if this start is OK
        if (type in ["firmware", "main", "event"]):
            if (len(self.current_sections) > 0):
                AsmError_NO_RET(117, "This section %s must be the first section" % (type))

            if (self.download_type):
                if (self.download_type[0] in ["main", "event"] and type in ["firmware"]):
                    AsmError_NO_RET(118, "Can't mix 'firmware' and 'main'/'event' sections")
                elif (self.download_type[0] == "firmware"):
                    AsmError_NO_RET(119, "Can't mix 'firmware' and 'main'/'event' sections " +
                                          "or have multiple 'firmware' sections")
                elif (type == "main" and 'main' in self.download_type):
                    AsmError_NO_RET(120, "Can't have multiple 'main' sections")
                self.download_type.append(type)
            else:
                self.download_type = [type]

        # add the section
        self.current_sections.append(type)
        self.section_breaks.append((type, self.stream_marker(), -1))
        self.section_args.append((arg1, arg2, arg3))
        self.section_count += 1

    def add_end(self, type):
        io.Out.DebugRaw("Add_end type:%s" % (type))

        # check if we are actually in the section?
        if (type not in self.current_sections):
            AsmError_NO_RET(121, "Not in section %s, so can't end it" % (type))

        # add the ending index into the beginning bit
        i = 0
        while i < len(self.section_breaks):
            s_type, start, end = self.section_breaks[i]
            if (s_type == type):
                self.section_breaks[i] = ((s_type, start, self.stream_marker()))
                break
            i += 1
        self.current_sections.remove(type)

    # Output function
    def dump_tokens(self, inc_src=False):
        print("Dumping tokens\n")
        i = 0
        for t in self.token_stream:
            # is this a start of a section?
            for name, start, end in self.section_breaks:
                if (start == i):
                    print(" ****** SECTION START " + name + " *****\n")
                    break

            # Is there a label for this index?
            for name in self.labels:
                if (self.labels[name] == i):
                    if (name.startswith(":")):
                        print("===== FUNCTION :" + name + " =====")
                    else:
                        print(":" + name)
                    break

            if (inc_src and t.source_line):
                print(t.source_line)
            print("%3d: " % (i), end='')
            t.print_token()
            print("")
            i += 1


class TokenAnalyser(object):
    def __init__(self, token_stream):
        self.token_stream = token_stream
        self.name_space_map = [{}, {}, {}]
        self.name_space_max = [0, 0, 0]

    def verify(self):
        return True

    def map_all_variables(self):
        io.Out.SetErrorRawContext(2, "Mapping all variables")
        self.name_space_map = [{}, {}, {}]
        for i in range(3):
            self.map_variables_in_space(i, self.name_space_map[i])

        # self.dump_variable_map()

        # They all fit, now fixup the tokens
        io.Out.SetErrorRawContext(2, "Fixing up variable references")
        for t in self.token_stream.token_stream:
            for index, space, name in t.var_info:
                if (name not in self.name_space_map[space]):
                    AsmError_NO_RET(122, "Variable %s not declared in %s space." % (name, space_names[space]))
                t.fixup_var_byte(index, self.name_space_map[space][name][0])

    def dump_variable_map(self):
        print("\nDumping variable maps:\n")
        for i in range(2):                    # skip LCD
            print("Space:", space_names[i])
            for name in self.name_space_map[i]:
                start, length = self.name_space_map[i][name]
                if (length == 1):
                    print("%10s at %d" % (name, start))
                else:
                    print("%10s at %d-%d" % (name, start, start + length - 1))

            print

    def map_variables_in_space(self, space, v_map):
        # create a map for the variables and fit them in
        io.Out.SetErrorRawContext(2, "Mapping variables in %s space" % (space_names[space]))
        limit = self.token_stream.limits[space]
        variables = self.token_stream.name_space[space]

        fixed_v_free = [(0, limit, limit)]

        # fit in the reserved and fixed variables
        for type, name, start, length in variables:
            if (type == 'data' and start < 0):
                # don't handle unfixed data here
                continue

            end = start + length
            copy_to_end = False
            v_free = fixed_v_free
            fixed_v_free = []

            # print("v_free", v_free)
            # print("v_map", v_map)

            for f_start, f_length, f_end in v_free:
                if (copy_to_end):
                    fixed_v_free.append((f_start, f_length, f_end))
                    continue

                if (start >= f_start and start < f_end):
                    # this is the block
                    if (end <= f_end):
                        # and it fits!
                        if (type == "data"):
                            if (name in v_map):
                                if (v_map[name][0] != start or v_map[name][1] != length):
                                    # the name is already used and the declarations are different!
                                    AsmError_NO_RET(123, "Data variable %s declared twice and differently!" % (name))
                            else:
                                v_map[name] = (start, length)

                        if (start == f_start):
                            if (end == f_end):
                                # just remove this entry from the free list
                                pass
                            else:
                                # remove the front end of the free block
                                fixed_v_free.append((end, f_length - length, f_end))

                        elif (end == f_end):
                            # remove the back end of the free block
                            fixed_v_free.append((f_start, f_length - length, start))
                        else:
                            # in the middle, create two smaller free blocks
                            fixed_v_free.append((f_start, start - f_start, start))
                            fixed_v_free.append((end, f_end - end, f_end))

                        # found the data, copy the rest to the end
                        copy_to_end = True

                    else:
                        if (type == "data"):
                            io.Out.ErrorRaw("Fixed data variable %s at %d didn't fit!" % (name, start))
                            io.Out.Error(io.TS.ASM_MEM_OVERFLOW,
                                         "file::: Overflowed {0} memory", "fixed")
                            raise program.AssemblerError
                        else:
                            io.Out.ErrorRaw("No room for Reserved data space at %d" % (start))
                            io.Out.Error(io.TS.ASM_MEM_OVERFLOW,
                                         "file::: Overflowed {0} memory", "rsvd")
                            raise program.AssemblerError

        # Now try to fit in the floating variables. Use a best-fit strategy where the smallest
        # hole is used for each request. Also the block sizes are tried largest to smallest
        floats = [(x[1], x[3]) for x in variables if x[0] == "data" if x[2] < 0]
        floats.sort(key=lambda x: x[1], reverse=True)
        v_free = fixed_v_free

        for name, length in floats:

            # find the best fit!
            i = 0
            best_index = -1
            for i in range(len(v_free)):
                f_length = v_free[i][1]
                if (f_length >= length):
                    if (f_length == length):
                        # perfect fit!
                        best_index = i
                        break

                    elif (best_index == -1):
                        best_index = i
                    else:
                        if (f_length < v_free[best_index][1]):
                            best_index = i

            # best_index == -1 if nothing fit
            if (best_index == -1):
                AsmError_NO_RET(124, "Float data variable %s (len:%d) didn't fit!" % (name, length))

            # Add to the map and adjust free spaces
            f_start, f_length, f_end = v_free[best_index]
            if (name in v_map):
                # the name is already used!
                AsmError_NO_RET(125, "Data variable %s declared twice!" % (name))
            else:
                v_map[name] = (f_start, length)

            if (length == f_length):
                # just remove this entry from the free list
                del v_free[best_index]

            else:
                # put it at the start
                v_free[best_index] = (f_start + length, f_length - length, f_end)

            # print("v_free", v_free)
            # print("v_map", v_map)

        # Write the max index for this space
        max = 0
        for n in v_map:
            maybe_max = v_map[n][0] + v_map[n][1]
            if (maybe_max > max):
                max = maybe_max

        self.name_space_max[space] = max

        return True

    def calc_cumulative_lengths(self, c_lengths):
        # i = 0
        # calculate cumulative lengths
        cumulative_length = 0
        for t in self.token_stream.token_stream:
            c_lengths.append(cumulative_length)
            cumulative_length += t.get_byte_len()
        c_lengths.append(cumulative_length)
        # print(c_lengths)

    def fixup_jumps(self):
        # fixup the jumps
        io.Out.SetErrorRawContext(2, "Fixing up jumps")
        # print(self.token_stream.labels)

        # First get a cumulative byte length for each index
        c_lengths = []
        self.calc_cumulative_lengths(c_lengths)

        # Verify that the labels exist
        for t in self.token_stream.token_stream:
            if (t.has_jump_label()):
                index, name, big = t.get_jump_label()
                if (name not in self.token_stream.labels):
                    AsmError_NO_RET(126, "Reference to an unknown label:%s" % (name))

        # do multiple passes until all jumps can be satisfied (start with byte)
        while (1):
            c_lengths = []
            self.calc_cumulative_lengths(c_lengths)

            #
            i = 0
            while (i < len(self.token_stream.token_stream)):
                t = self.token_stream.token_stream[i]
                if (t.has_jump_label()):
                    index, name, big = t.get_jump_label()
                    my_address = c_lengths[i + 1]  # The PC is pointing to start of next token
                    target_address = c_lengths[self.token_stream.labels[name]]
                    offset = target_address - my_address

                    # does it fit?
                    if (not big):
                        if ((offset > MAX_SBYTE) or (offset < MIN_SBYTE)):
                            # must become big now - will have to re-compute everyone
                            t.fixup_jump(True, offset)
                            break
                        else:
                            # small is OK
                            t.fixup_jump(False, offset)
                    else:
                        # keep it big as the size never gets smaller
                        t.fixup_jump(True, offset)
                i += 1

            if (i >= len(self.token_stream.token_stream)):
                # no more fixups needed
                break

        return True

    def get_max_location(self):
        keys = self.token_stream.devices.keys()
        if (keys):
            maxl = max(keys)
        else:
            maxl = -1
        return maxl

    def get_loc_type_and_size(self, loc):
        if (loc in self.token_stream.devices):
            return self.token_stream.devices[loc]
        else:
            return (0, 0)

    def create_header(self):
        # print(self.token_stream.download_type)
        io.Out.SetErrorRawContext(2, "Creating download header")
        if (not self.token_stream.version):
            AsmError_NO_RET(127, "Version wasn't set")

        else:
            if (self.token_stream.download_type[0] == "firmware"):
                if ((self.token_stream.version[0] != 0xe) and
                    (self.token_stream.version[0] != 0x2)):
                    AsmError_NO_RET(128, "This assembler only handles firmware " +
                                    "major versions 0xe or 0x2 (not 0x%x)" %
                                    (self.token_stream.version[0]))
            else:
                if (self.token_stream.version[0] != 0x6):
                    AsmError_NO_RET(129, "This assembler only handles token " +
                                    "major versions 0x6 (not 0x%x)" % (self.token_stream.version[0]))

        header_list = []
        added_bytes = 0
        if (self.token_stream.download_type[0] == "firmware"):
            # header is just size and crc
            bytes = []
            for t in self.token_stream.token_stream:
                bytes.extend(t.get_token_bits())

            # skip bad crc lengths
            if (is_bad_length_for_crc(len(bytes))):
                print("Warning - skipping bad CRC length at {} bytes.".format(len(bytes)))
                added_bytes = 1
                bytes.extend([0xff])

            header_list.extend([0, 0, 0, 0])
            header_list[0], header_list[1] = word_to_bytes(len(bytes))
            crc = calculate_crc(bytes)
            header_list[2], header_list[3] = word_to_bytes(crc)
            return ("firmware", self.token_stream.version, header_list, added_bytes)

        else:
            bytes = []
            for t in self.token_stream.token_stream:
                bytes.extend(t.get_token_bits())


            # # TEST CODE - get to a bad crc length
            # print("Token bytes:", len(bytes))
            # if (len(bytes) < 760):
            #     added_bytes = 760-len(bytes)
            #     print("Added", added_bytes, "bytes")
            #     bytes.extend([0xff] * added_bytes)
            #     print("Bytes now:", len(bytes))

            # data_bytes(2), data_crc(2), 8-bit vars, 16-bit vars, program_offset(2)
            header_list.extend([0, 0, 0, 0, 0, 0, 0, 0])
            event_list = []
            main_offset = 0

            # put in the name_space maxes
            header_list[4] = self.name_space_max[0]
            header_list[5] = self.name_space_max[1]

            # need locations for event and program offsets
            c_lengths = []
            self.calc_cumulative_lengths(c_lengths)

            # build the event list but first with offsets from start of code tokens
            for i in range(len(self.token_stream.section_breaks)):
                stype, start_token, stop_token = self.token_stream.section_breaks[i]
                if (stype == "main"):
                    main_offset = c_lengths[start_token]
                else:
                    modreg, mask, value = self.token_stream.section_args[i]
                    event_list.append((c_lengths[start_token], modreg, mask, value))

            final_header_bytes = len(header_list) + len(event_list) * 5 + 2

            # fixup main offset
            header_list[6], header_list[7] = word_to_bytes(main_offset + final_header_bytes)

            # append the event list to the current header
            for (offset, modreg, mask, value) in event_list:
                work = [0, 0, modreg, mask, value]
                work[0], work[1] = word_to_bytes(offset + final_header_bytes)
                header_list.extend(work)

            # mark end of events
            header_list.extend([0, 0])

            # finally update the length and crc
            new_bytes = header_list[4:] + bytes

            # skip bad crc lengths
            if (is_bad_length_for_crc(len(new_bytes))):
                print("Warning - skipping bad CRC length at {} bytes.".format(len(new_bytes)))
                added_bytes += 1
                new_bytes.extend([0xff])

            header_list[0], header_list[1] = word_to_bytes(len(new_bytes))
            crc = calculate_crc(new_bytes)
            header_list[2], header_list[3] = word_to_bytes(crc)

            return ("program", self.token_stream.version, header_list, added_bytes)

    def dump_extras(self):
        print("Section breaks:", self.token_stream.section_breaks)
        print("Jump labels:", self.token_stream.labels)


class Token(object):
    def __init__(self, type, err_reporter=None, src=None):
        self.valid = True
        self.token_info = []
        self.var_info = []
        self.jump_label = None
        self.type = type
        self.cached_bits = []
        self.binary_file = None

        if (src and src.endswith('\n')):
            self.source_line = src[:-1]
        else:
            self.source_line = src

    def mark_invalid(self):
        self.valid = False

    def invalidate_cache(self):
        self.cached_bits = []

    def get_type(self):
        return self.type

    def add_byte(self, index, value):
        # print("DEBUG - add_byte() - index:%s, value:%s" % (index, value))
        if (value < MIN_BYTE or value > MAX_BYTE):
            self.mark_invalid()
            AsmError_NO_RET(130, "Out of range for a byte: %d" % (value))
        else:
            self.token_info.append((index, 0, 0xff, value))

        self.invalidate_cache()

    def find_index(self, index):
        found = -1
        for i in range(len(self.token_info)):
            if (self.token_info[i][0] == index):
                found = i
                break
        if (found == -1):
            self.mark_invalid()
            AsmError_NO_RET(131, "Variable index: %d invalid" % (index))
        return found

    def fixup_crc(self, size, index, value):
        if (size == 8):
            if (value < 0 or value > MAX_BYTE):
                self.mark_invalid()
                AsmError_NO_RET(132, "Out of range for byte: %d" % (value))
            else:
                del self.token_info[self.find_index(index)]
                self.token_info.append((index, 0, 0xff, value))

        else:
            if (value < 0 or value > MAX_UWORD):
                self.mark_invalid()
                AsmError_NO_RET(133, "Out of range for an unsigned word: %d" % (value))
            else:
                del self.token_info[self.find_index(index)]
                del self.token_info[self.find_index(index + 1)]
                self.token_info.append((index, 0, 0xff, (value >> 8) & 255))
                self.token_info.append((index + 1, 0, 0xff, value & 255))

        self.invalidate_cache()

    def fixup_var_byte(self, index, value):
        i = self.find_index(index)
        new_number = self.token_info[i][3] + value
        if (new_number < MIN_BYTE or new_number > MAX_BYTE):
            self.mark_invalid()
            AsmError_NO_RET(134, "Out of range for a byte: %d" % (new_number))

        self.token_info[i] = (index, 0, 0xff, new_number)
        self.invalidate_cache()

    def fixup_jump(self, big, offset):
        j_index, j_name, j_big = self.jump_label
        if (big != j_big):
            if (not big):
                self.mark_invalid()
                AsmError_NO_RET(135, "Impossible - the jump size got SMALLER")

            # make the small into big
            del self.token_info[self.find_index(j_index)]
            self.add_bits(0, 4, 1, 1)
            self.add_word(j_index, offset)
            self.jump_label = (j_index, j_name, big)

        elif (big):
            # update the big one
            del self.token_info[self.find_index(j_index)]
            del self.token_info[self.find_index(j_index + 1)]
            self.add_word(j_index, offset)
        else:
            # update the small one
            del self.token_info[self.find_index(j_index)]
            if (offset < 0):
                # convert to a signed offset
                offset += 256
            self.add_byte(j_index, offset)
            self.token_info.append((j_index, 0, 0xff, offset))

        self.invalidate_cache()

    def add_word(self, index, value):
        if (value < MIN_WORD or value > MAX_WORD):
            self.mark_invalid()
            AsmError_NO_RET(136, "Out of range for a word: %d" % (value))
        else:
            self.token_info.append((index, 0, 0xff, (value >> 8) & 255))
            self.token_info.append((index + 1, 0, 0xff, value & 255))

        self.invalidate_cache()

    def add_uword(self, index, value):
        if (value < 0 or value > MAX_UWORD):
            self.mark_invalid()
            AsmError_NO_RET(137, "Out of range for an unsigned word: %d" % (value))
        else:
            self.token_info.append((index, 0, 0xff, (value >> 8) & 255))
            self.token_info.append((index + 1, 0, 0xff, value & 255))

        self.invalidate_cache()

    def add_bits(self, index, shift, mask, value):
        self.token_info.append((index, shift, mask, value))
        self.invalidate_cache()

    # Adds a name reference to the token[index]. The address of the name is added to the
    # value of the byte at that index.
    def add_vname(self, index, space, name):
        self.var_info.append((index, space, name))
        self.invalidate_cache()

    def clear_vnames(self):
        self.var_info = []

    def set_jump_label(self, index, name, big=False):
        # start off with a small jump
        self.jump_label = (index, name, big)
        self.invalidate_cache()

    def get_jump_label(self):
        return self.jump_label

    def has_jump_label(self):
        return self.jump_label is not None

    def clear_jump_label(self):
        self.jump_label = None

    def finish(self, token_stream, explicit_placement=-1):
        # add the token to the token stream for post-processing
        if (self.valid):
            token_stream.add_token(self, explicit_placement)
        else:
            io.Out.DebugRaw("Dropping invalid token")

    def get_byte_len(self):
        if (len(self.cached_bits) == 0):
            self.get_token_bits()

        return len(self.cached_bits)

    def get_token_bits(self):
        if (len(self.cached_bits) == 0):
            if (self.binary_file):
                fh = file(self.binary_file, 'rb')
                tmp_ascii = fh.read()
                for t in tmp_ascii:
                    self.cached_bits.append(ord(t))

                fh.close()
            else:
                length = 1
                self.cached_bits = [0]
                for index, shift, mask, value in self.token_info:
                    while (index + 1 > length):
                        self.cached_bits.append(0)
                        length += 1

                    work = self.cached_bits[index]
                    work = work & ~(mask << shift)
                    work = work | ((value & mask) << shift)
                    self.cached_bits[index] = work

        return self.cached_bits

    def add_binary_file(self, f_name):
        # save the file_name and bring it out when we get length or bits
        self.binary_file = f_name

    def print_token(self):
        length = self.get_byte_len()
        bits = self.get_token_bits()

        if (self.binary_file):
            print("Inserted %d binary bytes from %s:\n   [" % (length, self.binary_file), end='')

        for i in range(length):
            print("%02x " % (bits[i]), end='')
            if (((i + 1) % 20) == 0):
                print("\n    ")

        if (self.binary_file):
            print("]")

        elif (self.var_info):
            print("(var_refs:", end='')
            info = ""
            for i, s, n in self.var_info:
                if (info):
                    info += ", "
                info += "%s/%d/%s" % (n, i, space_names[s])
            print("%s)" % (info))

        elif (self.jump_label):
            print("(jump_ref to %s, big? %d)" % (self.jump_label[1], self.jump_label[2]))

        else:
            print()



#!/usr/bin/env python

# * **************************************************************** **
#
# File: EdAsm.py
# Desc: Front-end for the token assembler
# Note:
#
# Author: Brian Danilko, Likeable Software (brian@likeablesoftware.com)
#
# Copyright 2006, 2014, 2016, 2017 Microbric Pty Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License (in the doc/licenses directory)
# for more details.
#
# * **************************************************************** */

from __future__ import print_function
from __future__ import absolute_import

import sys
import argparse

from lib import token_assembler
from lib import hl_parser
from lib import io
from lib import util
from lib import audio


def assemble(options):

    download_bytes, download_str, download_type, version = \
        token_assembler.assemble_file(options.srcPath, options.debug)

    if (len(download_bytes) == 0):
        print("ERROR - No output produced")
        return

    # some sort of output
    print("Assembly completed of %s type file: %s -- created %d bytes of tokens and header" %
          (download_type, options.srcPath, len(download_str)))

    versionNumber = (version[0] << 4) + version[1]
    versionString = chr(versionNumber) + chr(255 - versionNumber)
    # print(versionNumber, ord(versionString[0]), ord(versionString[1]), download_type)

    full_download_str = versionString + download_str
    full_download_bytes = [versionNumber, 255 - versionNumber]
    full_download_bytes.extend(download_bytes)
    print(len(full_download_bytes))

    if (options.binFile is not None):
        if (options.preamble):
            download_str = full_download_str
        print("Writing %d bytes to file: %s" % (len(download_str), options.binFile.name))
        options.binFile.write(download_str)
        options.binFile.close()

    if (options.wavFilename is not None):
        audioFile = audio.Output(".", options.wavFilename)
        audioFile.WriteWav(full_download_bytes)


def ProcessCommandArgs(args):
    """Handle the command args and display usage if needed.
       Note that the usage is in English as we don't necessarily
       have the language file to use. Also, this will be run in
       the server, which shouldn't make mistakes."""

    # shenanigans to ensure order of the elements for help text
    levelChoices = (("error", io.LEVEL.ERROR), ("top", io.LEVEL.TOP),
                    ("warn", io.LEVEL.WARN), ("info", io.LEVEL.INFO),
                    ("verbose", io.LEVEL.VERBOSE,), ("debug", io.LEVEL.DEBUG))

    version = "1.2.11"
    parser = argparse.ArgumentParser(prog="EdAsm.py",
                                     description="Token assembler for Edison and Ed.Py, version %s" % (version,))

    parser.add_argument("-v", action="version", version="%(prog)s " + version)

    parser.add_argument("-r", "--reghelp", action="store_true", dest="reghelp",
                        help="Output all of the device types, locations and registers")

    parser.add_argument("-d", "--debug", action="store_true", dest="debug",
                        help="Output parsing information.")

    parser.add_argument("-s", "--source", action="store_true", dest="source",
                        help="Output source info with debug info.")

    parser.add_argument("-w", "--wav",
                        dest="wavFilename", default=None,
                        help="Output a wav file of the assembled code.")

    parser.add_argument("-b", "--bin", type=argparse.FileType('wb', 0),
                        dest="binFile", default=None,
                        help="Output a binary file of the assembled code.")

    parser.add_argument("-p", "--preamble", action="store_true", dest="preamble",
                        help="Add preamble to the binary file written.")

    parser.add_argument("-l", type=util.LowerStr,
                        choices=list(zip(*levelChoices))[0], default="error",  # default="debug",
                        help="Output level (default:%(default)s). " +
                        "\nAll output from previous levels and this one will be generated")

    parser.add_argument("srcPath", metavar="SRC", nargs='?', default=None,
                        help="Path to the source to be assembled")

    parsed = parser.parse_args(args)

    outputLevel = [x[1] for x in levelChoices if x[0] == parsed.l][0]

    if (outputLevel >= 0):
        # setup output
        io.Out.SetSink(io.SINK.CONSOLE)
        io.Out.SetMaxLevel(outputLevel)
        # io.Out.SetReRaise(parsed.reraise)

    if (parsed.source and not parsed.debug):
        print("Warning: --source (or -s) makes no sense without --debug (or -d).")

    if (parsed.preamble and parsed.binFile is None):
        print("Warning: --preamble (or -p) makes no sense without --bin (or -b).")

    if (parsed.reghelp):
        print("Device type, locations and register help")
        print("----------------------------------------")
        hl_parser.dump_reg_help()
        sys.exit(0)

    if (parsed.srcPath is None):
        print("Error -- missing source path")
        sys.exit(0)

    return parsed


def main(args):
    # clear global memory
    hl_parser.reset_devices_and_locations()
    token_assembler.reset_tokens()

    pargs = ProcessCommandArgs(args)

    assemble(pargs)


#####################################
if __name__ == "__main__":
    main(sys.argv[1:])
